library flip_renderer;

import 'dart:html';
import 'dart:typed_data';
import 'dart:web_gl';

const String _VertexShader = """
attribute vec4 position;
attribute vec2 tex_coord;
varying vec2 uv;
void main(void)
{
    gl_Position = position;
    uv = tex_coord;
}
""";
const String _FragmentShader = """
precision mediump float;
uniform sampler2D texture;
uniform float percent;
uniform float tilt;
uniform vec2 size;
uniform float roll_size;
varying vec2 uv;

float px(float uvx) {
    return uvx * size.x;
}

float py(float uvy) {
    return uvy * size.y;
}

vec2 to_uv(vec2 pos) {
    return vec2({l2r} pos.x / size.x, 1.0 - pos.y / size.y);
}

void main()
{
    float x1 = px({l2r} uv.x);
    float y1 = py(1.0 - uv.y);
    float per = px(percent);
    if (y1 / tilt + per < x1) {
        gl_FragColor = vec4(0.0, 0.0, 0.0, 0.0);
    } else {
        float x0 = (x1 / tilt + y1 + per * tilt) / (tilt + 1.0/tilt);
        float off = x1 - x0;
        float dis = abs(off);
        float x2 = 2.0 * x0 - x1;
        float y2 = 2.0 * (x1 - x0) / tilt + y1;
        if (off > -roll_size && off <= 0.0) {
            y1 = y1 - (-sqrt(-off * off - 2.0 * roll_size * off) + roll_size);
        }
        if (y1 > 0.0) {
            off = x2 - x0;
            if (off > 0.0) {
                if (off < roll_size) {
                    y2 = y2 - (sqrt(-off * off + 2.0 * roll_size * off) + roll_size);
                } else {
                    y2 = y2 - 2.0 * roll_size;
                }
            }
            if (y2 > 0.0 && y2 < size.y && x2 < size.x) {
                vec4 shadow = mix(vec4(0.6, 0.6, 0.6, 1.0), vec4(0.98, 0.98, 0.98, 1.0), min(1.0, dis/20.0));
                gl_FragColor = texture2D(texture, to_uv(vec2(x2, y2))) * shadow;
            } else {
                vec4 shadow = mix(vec4(1.2, 1.2, 1.2, 1.0), vec4(1.0, 1.0, 1.0, 1.0), min(1.0, dis/20.0));
                float sha_off = pow(max(0.0, max(y2 - size.y, -y2)), 2.0) + pow(max(0.0, x2 - size.x), 2.0);
                shadow = shadow - (1.0 - smoothstep(0.0, 81.0, sha_off)) * vec4(0.3, 0.3, 0.3, 0.0);
                gl_FragColor = texture2D(texture, to_uv(vec2(x1, y1))) * shadow;
            }
        } else {
            gl_FragColor = vec4(0.0, 0.0, 0.0, 0.0);
        }
    }
}
""";

class FlipRenderer {
  final RenderingContext context;

  Program? loadProgram(String vertex, String fragment) {
    var vertexShader = loadShader(WebGL.VERTEX_SHADER, vertex);
    if (vertexShader == null) return null;

    var fragmentShader = loadShader(WebGL.FRAGMENT_SHADER, fragment);
    if (fragmentShader == null) {
      context.deleteShader(vertexShader);
      return null;
    }

    // Create the program object
    var programHandle = context.createProgram();

    context.attachShader(programHandle, vertexShader);
    context.attachShader(programHandle, fragmentShader);

    // Link the program
    context.linkProgram(programHandle);

    context.deleteShader(vertexShader);
    context.deleteShader(fragmentShader);

    return programHandle;
  }

  Shader? loadShader(int type, String shaderStr) {
    Shader shader = context.createShader(type);

    context.shaderSource(shader, shaderStr);

    // Compile the shader
    context.compileShader(shader);

    var ret = context.getShaderParameter(shader, WebGL.COMPILE_STATUS);
    if (ret == false) {
      var log = context.getShaderInfoLog(shader);
      print("FlipTexture: $log");
      return null;
    }

    return shader;
  }

  late Program _programHandle;
  int _positionAttr = 0;
  int _texCoordAttr = 0;
  late UniformLocation _textureUniform;
  late UniformLocation _percentUniform;
  late UniformLocation _tiltUniform;
  late UniformLocation _sizeUniform;
  late UniformLocation _rollSizeUniform;

  Texture? _mainTexture;
  int _imageWidth = 0;
  int _imageHeight = 0;
  double rollSize;

  final int textureWidth;
  final int textureHeight;

  final bool leftToRight;

  late Buffer _posBuffer;
  late Buffer _uvBuffer;

  void initialize() {
    var programHandle = loadProgram(_VertexShader,
        _FragmentShader.replaceAll("{l2r}", leftToRight ? "1.0 - " : ""));
    if (programHandle == null) return;
    _programHandle = programHandle;

    var ret = context.getProgramParameter(_programHandle, WebGL.LINK_STATUS);
    if (ret == 0) {
      var log = context.getProgramInfoLog(_programHandle);
      print("FlipTexture:$log");
    }

    context.useProgram(_programHandle);
    _positionAttr = context.getAttribLocation(_programHandle, "position");
    _texCoordAttr = context.getAttribLocation(_programHandle, "tex_coord");
    _textureUniform = context.getUniformLocation(_programHandle, "texture");
    _percentUniform = context.getUniformLocation(_programHandle, "percent");
    _tiltUniform = context.getUniformLocation(_programHandle, "tilt");
    _sizeUniform = context.getUniformLocation(_programHandle, "size");
    _rollSizeUniform = context.getUniformLocation(_programHandle, "roll_size");

    Float32List pos = Float32List.fromList([
      -1.0,
      1.0,
      0.0,
      1.0,
      1.0,
      0.0,
      -1.0,
      -1.0,
      0.0,
      1.0,
      -1.0,
      0.0,
    ]);

    Float32List uv = Float32List.fromList([
      0,
      0,
      1,
      0,
      0,
      1,
      1,
      1,
    ]);

    // VBO will cause crash on android simulator
    _posBuffer = context.createBuffer();
    context.bindBuffer(WebGL.ARRAY_BUFFER, _posBuffer);
    context.bufferData(WebGL.ARRAY_BUFFER, pos, WebGL.STATIC_DRAW);

    _uvBuffer = context.createBuffer();
    context.bindBuffer(WebGL.ARRAY_BUFFER, _uvBuffer);
    context.bufferData(WebGL.ARRAY_BUFFER, uv, WebGL.STATIC_DRAW);
  }

  void updateTexture(int width, int height, Uint8List bytes) {
    _imageWidth = width;
    _imageHeight = height;
    if (_mainTexture != null) {
      context.deleteTexture(_mainTexture);
    }

    _mainTexture = context.createTexture();

    context.bindTexture(WebGL.TEXTURE_2D, _mainTexture);
    context.pixelStorei(WebGL.UNPACK_ALIGNMENT, 4);
    context.texParameteri(
        WebGL.TEXTURE_2D, WebGL.TEXTURE_MIN_FILTER, WebGL.LINEAR);
    context.texParameteri(
        WebGL.TEXTURE_2D, WebGL.TEXTURE_MAG_FILTER, WebGL.LINEAR);
    context.texParameteri(
        WebGL.TEXTURE_2D, WebGL.TEXTURE_WRAP_S, WebGL.CLAMP_TO_EDGE);
    context.texParameteri(
        WebGL.TEXTURE_2D, WebGL.TEXTURE_WRAP_T, WebGL.CLAMP_TO_EDGE);

    context.texImage2D(WebGL.TEXTURE_2D, 0, WebGL.RGBA, width, height, 0,
        WebGL.RGBA, WebGL.UNSIGNED_BYTE, bytes);
  }

  void draw(double percent, double tilt) {
    context.blendFunc(WebGL.ONE, WebGL.ONE_MINUS_SRC_ALPHA);
    context.clearColor(0, 0, 0, 0);
    context.clear(WebGL.COLOR_BUFFER_BIT);

    context.viewport(0, 0, textureWidth, textureHeight);

    context.useProgram(_programHandle);

    context.enableVertexAttribArray(_positionAttr);
    context.bindBuffer(WebGL.ARRAY_BUFFER, _posBuffer);
    context.vertexAttribPointer(_positionAttr, 3, WebGL.FLOAT, false, 0, 0);

    context.enableVertexAttribArray(_texCoordAttr);
    context.bindBuffer(WebGL.ARRAY_BUFFER, _uvBuffer);
    context.vertexAttribPointer(_texCoordAttr, 2, WebGL.FLOAT, false, 0, 0);

    context.activeTexture(WebGL.TEXTURE0);
    context.bindTexture(WebGL.TEXTURE_2D, _mainTexture);
    context.uniform1i(_textureUniform, 0);

    context.uniform1f(_percentUniform, percent);
    context.uniform1f(_tiltUniform, tilt);

    context.uniform2f(_sizeUniform, _imageWidth, _imageHeight);

    context.uniform1f(_rollSizeUniform, rollSize);

    context.drawArrays(WebGL.TRIANGLE_STRIP, 0, 4);
  }

  void destroy() {
    context.deleteBuffer(_posBuffer);
    context.deleteBuffer(_uvBuffer);
    context.deleteTexture(_mainTexture);
    context.deleteProgram(_programHandle);
  }

  FlipRenderer({
    required this.textureWidth,
    required this.textureHeight,
    required this.leftToRight,
    required this.context,
    rollSize = 12,
  });
}
